#!/usr/bin/env python2
from __future__ import print_function

"""filedrop.py: http upload target"""
__version__   = "1.1"
__author__    = "ed <filedrop@ocv.me>"
__url__       = "https://ocv.me/dev/?filedrop.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2019

# for a smaller (and much slower) alternative,
#   see https://ocv.me/dev/?filesink.py
#
# expect 492 MiB/s with python3, or 514 MiB/s using python2,
# that's like 16x faster than the built-in http parser at 32 MB/s
# (which is what filesink.py uses, cgi.FieldStorage)
#
# navigate to your ip:port combination with a browser to see a
# basic upload form, or just use curl to upload windows7.iso:
#   curl -F "f=@windows7.iso" http://127.0.0.1:1234/
#
# the http response is a minimal html which contains the
# filesize and sha1 checksum of the uploaded file, as seen by us
#
# 2019-05-13, 00:27, v1.1
#   support multiple files (slowdown from 546 to 514 MiB/s, 94%)
#
# 2019-05-01, 17:29, v1.0
#   initial working release

import os
import sys
import time
import hashlib
import select
import socket
import logging

logging.basicConfig(
	level = logging.DEBUG,
	format = '\033[36m%(asctime)s \033[33m%(levelname)5s \033[0m%(message)s'
)

PY2 = (sys.version_info[0] == 2)
if not PY2:
	unicode = str


##
## DEBUG HELPERS

from inspect import currentframe

HEX_WIDTH = 32

def _lno():
	return currentframe().f_back.f_lineno

def b2hex(data):
	if PY2:
		return ' '.join(map(lambda b: format(ord(b), "02x"), data))
	else:
		if type(data) is str:
			return ' '.join(map(lambda b: format(ord(b), "02x"), data))
		else:
			return ' '.join(map(lambda b: format(b, "02x"), data))

def hexdump(pk, prefix='', file=None):
	if file is not None:
		line_fmt = u'{0} {2}{3}{4}'
		hex_width = 4
		blk_width = 4
	else:
		line_fmt = u'{0}{1:8x}  {2}{3} {4}'
		hex_width = HEX_WIDTH
		blk_width = 8

	lpk = len(pk)
	ofs = 0
	hexofs = 0
	hexlen = 0
	hexstr = ''
	ascstr = ''
	ascstr_width = int(hex_width * 100 / 32.0 - 0.5)  # 32h = 100a, 16h = 50a
	while ofs < lpk:
		hexstr += b2hex(pk[ofs:ofs+blk_width])
		hexstr += ' '
		if PY2:
			ascstr += ''.join(map(lambda b: b if ord(b) >= 0x20 and ord(b) < 0x7f else '.', pk[ofs:ofs+blk_width]))
		else:
			ascstr += ''.join(map(lambda b: chr(b) if b >= 0x20 and b < 0x7f else '.', pk[ofs:ofs+blk_width]))
		hexlen += blk_width
		ofs += blk_width
		
		if hexlen >= hex_width or ofs >= lpk:
			txt = line_fmt.format(prefix, hexofs, hexstr,
				u' '*(ascstr_width-len(hexstr)), ascstr)
			
			if file is not None:
				file.write((txt + u'\n').encode('utf-8'))
			else:
				print(txt)

			hexofs = ofs
			hexstr = ''
			hexlen = 0
			ascstr = ''
		else:
			hexstr += ' '
			ascstr += ' '

## END DEBUG HELPERS
##


class Unrecv(object):
	def __init__(self, s):
		self.s = s
		self.buf = b''
	
	def recv(self, nbytes):
		if self.buf:
			ret = self.buf[:nbytes]
			self.buf = self.buf[nbytes:]
			return ret
		
		try:
			return self.s.recv(nbytes)
		except:
			return b''
	
	def unrecv(self, buf):
		self.buf = buf + self.buf


class UnbufferedHttpd(object):
	def __init__(self, addr, port):
		srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		srv.bind((addr, port))
		srv.listen(10)
		
		while True:
			UnbufferedClient(*srv.accept()).run()


class UnbufferedClient(object):
	def __init__(self, sck, addr):
		self.s = sck
		self.sr = Unrecv(sck)
		self.addr = addr
		self.ok = True
	
	def run(self):
		headerlines = self.read_header()
		if not self.ok:
			return
		
		self.headers = {}
		mode, self.req, _ = headerlines[0].split(' ')
		
		for header_line in headerlines[1:]:
			k, v = header_line.split(':', 1)
			self.headers[k.lower()] = v.strip()
		
		if mode == 'GET':
			self.handle_get()
		elif mode == 'POST':
			self.handle_post()
		else:
			self.loud_reply(u'invalid HTTP mode "{0}"'.format(mode))
	
	def panic(self, msg):
		logging.error('client disconnected ({0})'.format(msg).upper())
		self.ok = False
		self.s.close()
	
	def read_header(self):
		ret = b''
		while True:
			if ret.endswith(b'\r\n\r\n'):
				break
			elif ret.endswith(b'\r\n\r'):
				n = 1
			elif ret.endswith(b'\r\n'):
				n = 2
			elif ret.endswith(b'\r'):
				n = 3
			else:
				n = 4
			
			buf = self.sr.recv(n)
			if not buf:
				self.panic('headers')
				break
			
			ret += buf
		
		return ret[:-4].decode('utf-8', 'replace').split('\r\n')
	
	def reply(self, body):
		header = 'HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: {0}\r\n\r\n'.format(len(body)).encode('utf-8')
		if self.ok:
			self.s.send(header + body)
		
		self.s.close()
		return body
	
	def loud_reply(self, body):
		logging.info(body.rstrip())
		self.reply(b'<pre>' + body.encode('utf-8'))
	
	def handle_get(self):
		logging.info('')
		logging.info('GET  {0} {1}'.format(self.addr[0], self.req))
		self.reply(b'<form method="post" enctype="multipart/form-data"><h5><input type="file" name="f" multiple><h5><input type="submit" value="start upload">')
	
	def handle_post(self):
		logging.info('')
		logging.info('POST {0} {1}'.format(self.addr[0], self.req))
		
		try:
			if self.headers['expect'].lower() == '100-continue':
				self.s.send(b'HTTP/1.1 100 Continue\r\n\r\n')
		except:
			pass
		
		form_segm = self.read_header()
		if not self.ok:
			return
		
		self.boundary = b'\r\n' + form_segm[0].encode('utf-8')
		for ln in form_segm[1:]:
			logging.debug(ln)
		
		fn0 = 'inc.{0:.6f}'.format(time.time())
		#fn0 = '/dev/null'

		files = []
		t0 = time.time()
		for nfile in range(99):
			fn = fn0
			if fn0 != '/dev/null':
				fn = '{0}.{1}'.format(fn0, nfile)
			
			with open(fn, 'wb') as f:
				logging.info('writing to {0}'.format(fn))
				sz, sha1 = self.handle_multipart(f)
				if sz == 0:
					break
				
				files.append([sz, sha1])
			
			buf = self.sr.recv(2)
			
			if buf == b'--':
				# end of multipart
				break
			
			if buf != b'\r\n':
				return self.loud_reply(u'protocol error')
			
			header = self.read_header()
			if not self.ok:
				break
			
			form_segm += header
			for ln in header:
				logging.debug(ln)
		
		td = time.time() - t0
		sz_total = sum(x[0] for x in files)
		spd = (sz_total / td) / (1024*1024)
		
		status = 'OK'
		if not self.ok:
			status = 'ERROR'
		
		msg = u'{0} // {1} bytes // {2:.3f} MiB/s\n'.format(status, sz_total, spd)
		
		for sz, sha1 in files:
			msg += u'sha1: {0} // {1} bytes\n'.format(sha1, sz)
		
		self.loud_reply(msg)

		if fn != '/dev/null':
			with open(fn0 + '.txt', 'wb') as f:
				f.write((u'\n'.join(unicode(x) for x in [
					u':'.join(unicode(x) for x in self.addr),
					u'\n'.join(form_segm),
					msg.rstrip()
					]) + '\n').encode('utf-8'))
	
	def handle_multipart(self, ofd):
		tlen = 0
		hashobj = hashlib.sha1()
		for buf in self.extract_filedata():
			tlen += len(buf)
			hashobj.update(buf)
			ofd.write(buf)
		
		return tlen, hashobj.hexdigest()
	
	def extract_filedata(self):
		blen = len(self.boundary)
		while True:
			buf = self.sr.recv(32768)
			if not buf:
				# abort: client disconnected
				self.panic('outer')
				return
			
			#hexdump(buf, _lno())
			
			while True:
				ofs = buf.find(self.boundary)
				if ofs != -1:
					self.sr.unrecv(buf[ofs + blen:])
					#hexdump(buf[:ofs], 'ret')
					#hexdump(self.sr.buf, 'rem')
					yield buf[:ofs]
					return
				
				d = len(buf) - blen
				if d > 0:
					# buffer growing large; yield everything except
					# the part at the end (maybe start of boundary)
					yield buf[:d]
					buf = buf[d:]
				
				# look for boundary near the end of the buffer
				for n in range(1, len(buf)+1):
					if not buf[-n:] in self.boundary:
						n -= 1
						break
				
				if n == 0 \
				or not self.boundary.startswith(buf[-n:]):
					# no boundary contents near the buffer edge
					break
				
				if blen == n:
					# EOF: found boundary
					yield buf[:-n]
					return
				
				buf2 = self.sr.recv(32768)
				if not buf2:
					# abort: client disconnected
					self.panic('inner')
					return
				
				buf += buf2
			
			yield buf


port = 1234
if len(sys.argv) > 1:
	port = int(sys.argv[1])

logging.info('listening @ http://127.0.0.1:{0}/'.format(port))
UnbufferedHttpd('0.0.0.0', port)
